/*
 * Papillon - Solaris security module -  http://www.roqe.org/papillon
 * Copyright (c) 2000-2002 Konrad Rieck <kr@roqe.org>
 * $Id: stckbt.c,v 1.20 2002/09/02 09:33:41 kr Exp $
 *
 * This file contains everything related to the sticky bit on directories.
 * In other words the sym/hardlink and fifo protection.
 */


#define _KERNEL

#include "headers.h"

extern pap_syscall_t syscalls[];
#ifdef _64BIT
extern pap_syscall_t syscalls32[];
#endif

extern pap_config_t config;
extern krwlock_t config_rwlock;

#if defined(FIFOPROT) || defined(SYMPROT)
/*
 * Symlink and Fifo protection. 
 */
int stckbt_checks(char *path, int oflag)
{
    vnode_t *fvp, *pvp;		/* file and parent vnode */
    vattr_t fvattr, pvattr;	/* file and parent vnode attributes */
    cred_t *cred;
    boolean_t ret = B_TRUE;

    if (lookupname((char *) path, UIO_USERSPACE, NO_FOLLOW, &pvp, &fvp))
	return B_TRUE;

    if (!fvp || !pvp)
	goto err_and_release;

    if (VOP_GETATTR(fvp, &fvattr, 0, CRED()) ||
	VOP_GETATTR(pvp, &pvattr, 0, CRED()))
	goto err_and_release;

    cred = CRED();

#if defined(FIFOPROT)
    /*
     * Fifo protection
     */
    if ((oflag & O_CREAT) &&	/* create mode */
	cred->cr_ruid != SUSER_UID &&	/* not super-user */
	(pvattr.va_mode & S_ISVTX) &&	/* parent is sticky */
	fvp->v_type == VFIFO &&	/* file is fifo */
	fvattr.va_uid != cred->cr_uid &&	/* file uid not cred uid */
	pvattr.va_uid != fvattr.va_uid) {	/* file uid not parent uid */

	if (config.fifoprot == DENY) {
	    cmn_err(CE_WARN, "Denied opening fifo (owner %d:%d), process "
		    "%s (pid %d, uid %d)",
		    fvattr.va_uid, fvattr.va_gid, u.u_psargs,
		    ddi_get_pid(), cred->cr_ruid);
	    ret = B_FALSE;
	} else {
	    cmn_err(CE_WARN, "Opening fifo (owner %d:%d), process "
		    "%s (pid %d, uid %d)",
		    fvattr.va_uid, fvattr.va_gid, u.u_psargs,
		    ddi_get_pid(), cred->cr_ruid);
	}
    }
#endif				/* FIFOPROT */

#if defined(SYMPROT)
    /*
     * Symlink protection
     */
    if (ret && (pvattr.va_mode & S_ISVTX) &&	/* parent is sticky */
	fvp->v_type == VLNK &&	/* file is symlink */
	fvattr.va_uid != cred->cr_uid &&	/* file uid not cred uid */
	pvattr.va_uid != fvattr.va_uid) {	/* file uid not parent uid */


	if (config.symprot == DENY) {
	    cmn_err(CE_WARN,
		    "Denied following symlink (owner %d:%d), process "
		    "%s (pid %d, uid %d)", fvattr.va_uid, fvattr.va_gid,
		    u.u_psargs, ddi_get_pid(), cred->cr_ruid);
	    ret = B_FALSE;
	} else {
	    cmn_err(CE_WARN, "Following symlink (owner %d:%d), process "
		    "%s (pid %d, uid %d)",
		    fvattr.va_uid, fvattr.va_gid, u.u_psargs,
		    ddi_get_pid(), cred->cr_ruid);
	}
    }
#endif				/* SYMPROT */

  err_and_release:
    if (fvp)
	VN_RELE(fvp);
    if (pvp)
	VN_RELE(pvp);

    return ret;

}

/*
 * open64() redirection.
 * Perform fifo and symlink protections.
 */
int64_t pap_open64(const char *path, int oflag, mode_t mode)
{
    int64_t error;
    boolean_t access = B_TRUE;

    inc_refcnt();
    dcmn_err((CE_CONT, "open64() syscall.\n"));
    rw_enter(&config_rwlock, RW_READER);

    if (!config.symprot && !config.fifoprot)
	goto skip;

    access = stckbt_checks((char *) path, oflag);

  skip:
    if (access)
	error = syscalls[OPEN64].sc(path, oflag, mode);
    else
	error = set_errno(EPERM);

    rw_exit(&config_rwlock);
    dec_refcnt();

    return error;
}

/*
 * open() redirection.
 * Perform fifo and symlink protections.
 */
int64_t pap_open(const char *path, int oflag, mode_t mode)
{
    int64_t error;
    boolean_t access = B_TRUE;

    inc_refcnt();
    dcmn_err((CE_CONT, "open() syscall.\n"));
    rw_enter(&config_rwlock, RW_READER);

    if (!config.symprot && !config.fifoprot)
	goto skip;

    access = stckbt_checks((char *) path, oflag);

  skip:
    if (access)
	error = syscalls[OPEN].sc(path, oflag, mode);
    else
	error = set_errno(EPERM);

    rw_exit(&config_rwlock);
    dec_refcnt();

    return error;

}
#endif				/* SYM- & FIFOPROT */


#if defined(HARDPROT)
/*
 * link() redirection.
 * Perform hardlink protection.
 */
int64_t pap_link(const char *existing, const char *new)
{
    struct vnode *vp;
    struct vattr vattr;
    cred_t *cred;
    boolean_t access = B_TRUE;
    int64_t error;

    inc_refcnt();
    dcmn_err((CE_CONT, "link() syscall.\n"));
    rw_enter(&config_rwlock, RW_READER);

    if (!config.hardprot ||
	lookupname((char *) existing, UIO_USERSPACE, NO_FOLLOW, NULLVPP,
		   &vp))
	goto err;

    if (!vp)
	goto err_and_release;

    if (VOP_GETATTR(vp, &vattr, 0, CRED()))
	goto err_and_release;

    cred = CRED();

    /*
     * Hardlink protection.
     */
    if (cred->cr_uid != SUSER_UID &&	/* not super-user */
	cred->cr_uid != vattr.va_uid) {	/* file uid not cred uid */


	if (config.hardprot == DENY) {
	    cmn_err(CE_WARN, "Denied creating hardlink to (owner %d:%d), "
		    "process %s (pid %d, uid %d)",
		    vattr.va_uid, vattr.va_gid, u.u_psargs,
		    ddi_get_pid(), cred->cr_ruid);
	    access = B_FALSE;
	} else {
	    cmn_err(CE_WARN, "Creating hardlink to (owner %d:%d), "
		    "process %s (pid %d, uid %d)",
		    vattr.va_uid, vattr.va_gid, u.u_psargs,
		    ddi_get_pid(), cred->cr_ruid);
	}
    }

  err_and_release:
    if (vp)
	VN_RELE(vp);
  err:
    if (access)
	error = syscalls[LINK].sc(existing, new);
    else
	error = set_errno(EPERM);

    rw_exit(&config_rwlock);
    dec_refcnt();

    return error;
}

#endif				/* HARDPROT */

#ifdef _64BIT
#if defined(FIFOPROT) || defined(SYMPROT)
/*
 * open64() redirection.
 * Perform fifo and symlink protections.
 */
int64_t pap_open64_32(const char *path, int oflag, mode_t mode)
{
    int64_t error;
    boolean_t access = B_TRUE;

    inc_refcnt();
    dcmn_err((CE_CONT, "open64_32() syscall.\n"));
    rw_enter(&config_rwlock, RW_READER);

    if (!config.symprot && !config.fifoprot)
	goto skip;

    access = stckbt_checks((char *) path, oflag);

  skip:
    if (access)
	error = syscalls32[OPEN64].sc(path, oflag, mode);
    else
	error = set_errno(EPERM);

    rw_exit(&config_rwlock);
    dec_refcnt();

    return error;
}

/*
 * open() redirection.
 * Perform fifo and symlink protections.
 */
int64_t pap_open32(const char *path, int oflag, mode_t mode)
{
    int64_t error;
    boolean_t access = B_TRUE;

    inc_refcnt();
    dcmn_err((CE_CONT, "open32() syscall.\n"));
    rw_enter(&config_rwlock, RW_READER);

    if (!config.symprot && !config.fifoprot)
	goto skip;

    access = stckbt_checks((char *) path, oflag);

  skip:
    if (access)
	error = syscalls32[OPEN].sc(path, oflag, mode);
    else
	error = set_errno(EPERM);

    rw_exit(&config_rwlock);
    dec_refcnt();

    return error;

}
#endif				/* SYM- & FIFOPROT */
#endif				/* _64BIT */
